@@ -32,6 +32,40 @@ def getPassphrase(dbPath, verbose=True, infoFile=sys.stdout):
 						  verify=not dbExists)
 	return passphrase
 
+def run_infodump(dbPath):
+	try:
+		fc = libpwman.fileobj.FileObjCollection.parseFile(dbPath)
+		print("pwman database: %s" % dbPath)
+		head = fc.get(b"HEAD")
+		if head != libpwman.cryptsql.CryptSQL.CSQL_HEADER:
+			head = str(head)
+			if len(head) > 16:
+				head = head[:16] + "..."
+			raise libpwman.PWManError("Invalid HEAD: %s" % head)
+		for obj in fc.objects:
+			name = bytes(obj.getName())
+			data = bytes(obj.getData())
+			trunc = False
+			if name == b"PAYLOAD" and len(data) > 8:
+				data = data[:8]
+				trunc = True
+			try:
+				name = name.decode("UTF-8")
+			except UnicodeError as e:
+				raise libpwman.PWManError(
+					"Failed to decode file header name.")
+			try:
+				data = data.decode("UTF-8")
+			except UnicodeError as e:
+				data = data.hex()
+			if trunc:
+				data += "..."
+			pad = " " * (12 - len(name))
+			print("  %s%s: %s" % (name, pad, data))
+	except libpwman.fileobj.FileObjError as e:
+		raise libpwman.PWManError(str(e))
+	return 0
+
 def run_diff(dbPath, oldDbPath, diffFormat):
 	for p in (dbPath, oldDbPath):
 		if not p.exists():
@@ -54,7 +88,7 @@ def run_diff(dbPath, oldDbPath, diffFormat):
 		oldDb = libpwman.database.PWManDatabase(filename=oldDbPath,
 							passphrase=dbPassphrase,
 							readOnly=True)
-	except PWManError:
+	except libpwman.PWManError:
 		# The attempt failed. Ask the user for the proper passphrase.
 		dbPassphrase = getPassphrase(oldDbPath, verbose=True,
 					     infoFile=sys.stderr)
@@ -135,6 +169,13 @@ def run_ui(dbPath, timeout, commands):
 		return 1
 	return 0
 
+def runQuickSelfTests():
+	from libpwman.argon2 import Argon2
+	Argon2.get().quickSelfTest()
+
+	from libpwman.aes import AES
+	AES.get().quickSelfTest()
+
 def main():
 	p = argparse.ArgumentParser(
 		description="Commandline password manager - "
@@ -152,6 +193,8 @@ def main():
 	grp.add_argument("-c", "--command", action="append",
 			 help="Run this command instead of starting in interactive mode. "
 			      "-c|--command may be used multiple times.")
+	grp.add_argument("-I", "--info", action="store_true",
+			 help="Dump basic information about the database (without decrypting it).")
 	p.add_argument("-F", "--diff-format", type=lambda x: str(x).lower().strip(),
 		       default="unified",
 		       choices=("unified", "context", "ndiff", "html"),
@@ -161,7 +204,9 @@ def main():
 		       help="Use DB_PATH as database file. If not given, %s is used." % (
 			    libpwman.database.getDefaultDatabase()))
 	p.add_argument("--no-mlock", action="store_true",
-		       help="Do not lock memory and allow swapping to disk.")
+		       help="Do not lock memory and allow swapping to disk. "
+			    "Do not use this option, if you don't know what this means, "
+			    "because this option has security implications.")
 	if libpwman.util.osIsPosix:
 		p.add_argument("-t", "--timeout", type=int, default=600, metavar="SECONDS",
 			       help="Sets the session timeout in seconds. Default is 10 minutes.")
@@ -175,13 +220,12 @@ def main():
 	try:
 		interactiveMode = (not args.command and
 				   not args.diff and
-				   not args.call_pymod)
+				   not args.call_pymod and
+				   not args.info)
 
 		# Lock memory to RAM.
-		if not args.no_mlock:
-			MLockWrapper = libpwman.mlock.MLockWrapper
-			err = MLockWrapper.mlockall(MLockWrapper.MCL_CURRENT |
-						    MLockWrapper.MCL_FUTURE)
+		if not args.no_mlock and not args.info:
+			err = libpwman.mlock.MLockWrapper.get().mlockall()
 			baseMsg1 = "Failed to lock the pwman program memory to RAM to avoid "\
 				   "swapping secrets to disk.\nThe system call returned:"
 			baseMsg2 = "The contents of the decrypted password database "\
@@ -201,12 +245,13 @@ def main():
 							  "%s" % (
 							  err, baseMsg))
 
-		# Probe the AES singleton instance to bail out early
-		# in case of a crypto lib import error.
-		from libpwman.aes import AES
-		AES.get()
+		if not args.info:
+			runQuickSelfTests()
 
-		if args.diff:
+		if args.info:
+			assert not interactiveMode
+			exitcode = run_infodump(dbPath=args.database)
+		elif args.diff:
 			assert not interactiveMode
 			exitcode = run_diff(dbPath=args.database,
 					    oldDbPath=args.diff,
